@@ -1,30 +1,16 @@
-/*=============================================================================
-
-    This file is part of ARB.
-
-    ARB is free software; you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation; either version 2 of the License, or
-    (at your option) any later version.
-
-    ARB is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with ARB; if not, write to the Free Software
-    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
-
-=============================================================================*/
-/******************************************************************************
-
+/*
     Copyright (C) 2014 Fredrik Johansson
 
     2x2 mul code taken from MPFR 2.3.0
     (Copyright (C) 1991-2007 Free Software Foundation, Inc.)
 
-******************************************************************************/
+    This file is part of Arb.
+
+    Arb is free software: you can redistribute it and/or modify it under
+    the terms of the GNU Lesser General Public License (LGPL) as published
+    by the Free Software Foundation; either version 2.1 of the License, or
+    (at your option) any later version.  See <http://www.gnu.org/licenses/>.
+*/
 
 #ifndef ARF_H
 #define ARF_H
@@ -35,11 +21,18 @@
 #define ARF_INLINE static __inline__
 #endif
 
+#include <stdio.h>
 #include <math.h>
-#include "flint.h"
+#include "flint/flint.h"
 #include "fmpr.h"
 #include "mag.h"
 
+#ifndef flint_abort
+#if __FLINT_RELEASE <= 20502
+#define flint_abort abort
+#endif
+#endif
+
 #ifdef __cplusplus
 extern "C" {
 #endif
@@ -80,7 +73,7 @@ arf_rnd_to_mpfr(arf_rnd_t rnd)
 }
 
 /* Allow 'infinite' precision for operations where a result can be computed exactly. */
-#define ARF_PREC_EXACT LONG_MAX
+#define ARF_PREC_EXACT WORD_MAX
 
 #define ARF_PREC_ADD(prec,extra) ((prec) == ARF_PREC_EXACT ? ARF_PREC_EXACT : (prec) + (extra))
 
@@ -241,12 +234,7 @@ arf_init(arf_t x)
     ARF_XSIZE(x) = 0;
 }
 
-ARF_INLINE void
-arf_clear(arf_t x)
-{
-    fmpz_clear(ARF_EXPREF(x));
-    ARF_DEMOTE(x);
-}
+void arf_clear(arf_t x);
 
 ARF_INLINE void
 arf_zero(arf_t x)
@@ -361,6 +349,16 @@ int arf_cmp(const arf_t x, const arf_t y);
 
 int arf_cmpabs(const arf_t x, const arf_t y);
 
+int arf_cmpabs_ui(const arf_t x, ulong y);
+
+int arf_cmpabs_d(const arf_t x, double y);
+
+int arf_cmp_si(const arf_t x, slong y);
+
+int arf_cmp_ui(const arf_t x, ulong y);
+
+int arf_cmp_d(const arf_t x, double y);
+
 ARF_INLINE void
 arf_swap(arf_t y, arf_t x)
 {
@@ -441,10 +439,13 @@ arf_init_set_ui(arf_t x, ulong v)
     }
 }
 
+/* FLINT_ABS is unsafe for x = WORD_MIN */
+#define UI_ABS_SI(x) (((slong)(x) < 0) ? (-(ulong)(x)) : ((ulong)(x)))
+
 ARF_INLINE void
-arf_init_set_si(arf_t x, long v)
+arf_init_set_si(arf_t x, slong v)
 {
-    arf_init_set_ui(x, FLINT_ABS(v));
+    arf_init_set_ui(x, UI_ABS_SI(v));
     if (v < 0)
         ARF_NEG(x);
 }
@@ -471,21 +472,13 @@ arf_set_ui(arf_t x, ulong v)
 }
 
 ARF_INLINE void
-arf_set_si(arf_t x, long v)
+arf_set_si(arf_t x, slong v)
 {
-    arf_set_ui(x, FLINT_ABS(v));
+    arf_set_ui(x, UI_ABS_SI(v));
     if (v < 0)
         ARF_NEG(x);
 }
 
-ARF_INLINE int
-arf_cmpabs_ui(const arf_t x, ulong y)
-{
-    arf_t t;
-    arf_init_set_ui(t, y);  /* no need to free */
-    return arf_cmpabs(x, t);
-}
-
 ARF_INLINE void
 arf_init_set_shallow(arf_t z, const arf_t x)
 {
@@ -538,7 +531,7 @@ void arf_set_mpn(arf_t y, mp_srcptr x, mp_size_t xn, int sgnbit);
 ARF_INLINE void
 arf_set_mpz(arf_t y, const mpz_t x)
 {
-    long size = x->_mp_size;
+    slong size = x->_mp_size;
 
     if (size == 0)
         arf_zero(y);
@@ -555,34 +548,32 @@ arf_set_fmpz(arf_t y, const fmpz_t x)
         arf_set_mpz(y, COEFF_TO_PTR(*x));
 }
 
-int _arf_set_round_ui(arf_t x, ulong v, int sgnbit, long prec, arf_rnd_t rnd);
+int _arf_set_round_ui(arf_t x, ulong v, int sgnbit, slong prec, arf_rnd_t rnd);
 
-/* TODO: document */
-int _arf_set_round_uiui(arf_t z, long * fix, mp_limb_t hi, mp_limb_t lo, int sgnbit, long prec, arf_rnd_t rnd);
+int _arf_set_round_uiui(arf_t z, slong * fix, mp_limb_t hi, mp_limb_t lo, int sgnbit, slong prec, arf_rnd_t rnd);
 
-/* Assumes xn > 0, x[xn-1] != 0. */
 int
-_arf_set_round_mpn(arf_t y, long * exp_shift, mp_srcptr x, mp_size_t xn,
-    int sgnbit, long prec, arf_rnd_t rnd);
+_arf_set_round_mpn(arf_t y, slong * exp_shift, mp_srcptr x, mp_size_t xn,
+    int sgnbit, slong prec, arf_rnd_t rnd);
 
 ARF_INLINE int
-arf_set_round_ui(arf_t x, ulong v, long prec, arf_rnd_t rnd)
+arf_set_round_ui(arf_t x, ulong v, slong prec, arf_rnd_t rnd)
 {
     return _arf_set_round_ui(x, v, 0, prec, rnd);
 }
 
 ARF_INLINE int
-arf_set_round_si(arf_t x, long v, long prec, arf_rnd_t rnd)
+arf_set_round_si(arf_t x, slong v, slong prec, arf_rnd_t rnd)
 {
-    return _arf_set_round_ui(x, FLINT_ABS(v), v < 0, prec, rnd);
+    return _arf_set_round_ui(x, UI_ABS_SI(v), v < 0, prec, rnd);
 }
 
 ARF_INLINE int
-arf_set_round_mpz(arf_t y, const mpz_t x, long prec, arf_rnd_t rnd)
+arf_set_round_mpz(arf_t y, const mpz_t x, slong prec, arf_rnd_t rnd)
 {
     int inexact;
-    long size = x->_mp_size;
-    long fix;
+    slong size = x->_mp_size;
+    slong fix;
 
     if (size == 0)
     {
@@ -598,7 +589,7 @@ arf_set_round_mpz(arf_t y, const mpz_t x, long prec, arf_rnd_t rnd)
 }
 
 ARF_INLINE int
-arf_set_round_fmpz(arf_t y, const fmpz_t x, long prec, arf_rnd_t rnd)
+arf_set_round_fmpz(arf_t y, const fmpz_t x, slong prec, arf_rnd_t rnd)
 {
     if (!COEFF_IS_MPZ(*x))
         return arf_set_round_si(y, *x, prec, rnd);
@@ -606,9 +597,9 @@ arf_set_round_fmpz(arf_t y, const fmpz_t x, long prec, arf_rnd_t rnd)
         return arf_set_round_mpz(y, COEFF_TO_PTR(*x), prec, rnd);
 }
 
-int arf_set_round(arf_t y, const arf_t x, long prec, arf_rnd_t rnd);
+int arf_set_round(arf_t y, const arf_t x, slong prec, arf_rnd_t rnd);
 
-int arf_neg_round(arf_t y, const arf_t x, long prec, arf_rnd_t rnd);
+int arf_neg_round(arf_t y, const arf_t x, slong prec, arf_rnd_t rnd);
 
 void arf_get_fmpr(fmpr_t y, const arf_t x);
 
@@ -620,7 +611,7 @@ void arf_set_mpfr(arf_t x, const mpfr_t y);
 
 int arf_equal(const arf_t x, const arf_t y);
 
-int arf_equal_si(const arf_t x, long y);
+int arf_equal_si(const arf_t x, slong y);
 
 ARF_INLINE void
 arf_min(arf_t z, const arf_t a, const arf_t b)
@@ -649,7 +640,7 @@ arf_abs(arf_t y, const arf_t x)
         arf_set(y, x);
 }
 
-ARF_INLINE long
+ARF_INLINE slong
 arf_bits(const arf_t x)
 {
     if (arf_is_special(x))
@@ -658,7 +649,7 @@ arf_bits(const arf_t x)
     {
         mp_srcptr xp;
         mp_size_t xn;
-        long c;
+        slong c;
 
         ARF_GET_MPN_READONLY(xp, xn, x);
         count_trailing_zeros(c, xp[0]);
@@ -675,46 +666,16 @@ arf_bot(fmpz_t e, const arf_t x)
         fmpz_sub_si(e, ARF_EXPREF(x), arf_bits(x));
 }
 
-ARF_INLINE int
-arf_is_int(const arf_t x)
-{
-    if (arf_is_special(x))
-        return arf_is_zero(x);
-    else
-    {
-        fmpz_t t;
-        int r;
-        fmpz_init(t);
-        arf_bot(t, x);
-        r = fmpz_sgn(t) >= 0;
-        fmpz_clear(t);
-        return r;
-    }
-}
+int arf_is_int(const arf_t x);
 
-ARF_INLINE int
-arf_is_int_2exp_si(const arf_t x, long e)
-{
-    if (arf_is_special(x))
-        return arf_is_zero(x);
-    else
-    {
-        fmpz_t t;
-        int r;
-        fmpz_init(t);
-        arf_bot(t, x);
-        r = fmpz_cmp_si(t, e) >= 0;
-        fmpz_clear(t);
-        return r;
-    }
-}
+int arf_is_int_2exp_si(const arf_t x, slong e);
 
-int arf_cmp_2exp_si(const arf_t x, long e);
+int arf_cmp_2exp_si(const arf_t x, slong e);
 
-int arf_cmpabs_2exp_si(const arf_t x, long e);
+int arf_cmpabs_2exp_si(const arf_t x, slong e);
 
 ARF_INLINE void
-arf_set_si_2exp_si(arf_t x, long man, long exp)
+arf_set_si_2exp_si(arf_t x, slong man, slong exp)
 {
     arf_set_si(x, man);
     if (man != 0)
@@ -722,7 +683,7 @@ arf_set_si_2exp_si(arf_t x, long man, long exp)
 }
 
 ARF_INLINE void
-arf_set_ui_2exp_si(arf_t x, ulong man, long exp)
+arf_set_ui_2exp_si(arf_t x, ulong man, slong exp)
 {
     arf_set_ui(x, man);
     if (man != 0)
@@ -730,7 +691,7 @@ arf_set_ui_2exp_si(arf_t x, ulong man, long exp)
 }
 
 ARF_INLINE void
-arf_mul_2exp_si(arf_t y, const arf_t x, long e)
+arf_mul_2exp_si(arf_t y, const arf_t x, slong e)
 {
     arf_set(y, x);
     if (!arf_is_special(y))
@@ -746,7 +707,7 @@ arf_mul_2exp_fmpz(arf_t y, const arf_t x, const fmpz_t e)
 }
 
 ARF_INLINE int
-arf_set_round_fmpz_2exp(arf_t y, const fmpz_t x, const fmpz_t exp, long prec, arf_rnd_t rnd)
+arf_set_round_fmpz_2exp(arf_t y, const fmpz_t x, const fmpz_t exp, slong prec, arf_rnd_t rnd)
 {
     if (fmpz_is_zero(x))
     {
@@ -781,37 +742,24 @@ arf_abs_bound_le_2exp_fmpz(fmpz_t b, const arf_t x)
         fmpz_set(b, ARF_EXPREF(x));
 }
 
-long arf_abs_bound_lt_2exp_si(const arf_t x);
+slong arf_abs_bound_lt_2exp_si(const arf_t x);
+
+void arf_frexp(arf_t man, fmpz_t exp, const arf_t x);
 
 void arf_get_fmpz_2exp(fmpz_t man, fmpz_t exp, const arf_t x);
 
-void arf_get_fmpz(fmpz_t z, const arf_t x, arf_rnd_t rnd);
+int _arf_get_integer_mpn(mp_ptr y, mp_srcptr x, mp_size_t xn, slong exp);
 
-long arf_get_si(const arf_t x, arf_rnd_t rnd);
+int _arf_set_mpn_fixed(arf_t z, mp_srcptr xp, mp_size_t xn,
+        mp_size_t fixn, int negative, slong prec, arf_rnd_t rnd);
 
-ARF_INLINE int
-arf_get_fmpz_fixed_fmpz(fmpz_t y, const arf_t x, const fmpz_t e)
-{
-    int r;
-    fmpr_t t;
-    fmpr_init(t);
-    arf_get_fmpr(t, x);
-    r = fmpr_get_fmpz_fixed_fmpz(y, t, e);
-    fmpr_clear(t);
-    return r;
-}
+int arf_get_fmpz(fmpz_t z, const arf_t x, arf_rnd_t rnd);
 
-ARF_INLINE int
-arf_get_fmpz_fixed_si(fmpz_t y, const arf_t x, long e)
-{
-    int r;
-    fmpr_t t;
-    fmpr_init(t);
-    arf_get_fmpr(t, x);
-    r = fmpr_get_fmpz_fixed_si(y, t, e);
-    fmpr_clear(t);
-    return r;
-}
+slong arf_get_si(const arf_t x, arf_rnd_t rnd);
+
+int arf_get_fmpz_fixed_fmpz(fmpz_t y, const arf_t x, const fmpz_t e);
+
+int arf_get_fmpz_fixed_si(fmpz_t y, const arf_t x, slong e);
 
 ARF_INLINE void
 arf_set_fmpz_2exp(arf_t x, const fmpz_t man, const fmpz_t exp)
@@ -827,15 +775,27 @@ void arf_ceil(arf_t z, const arf_t x);
 
 void arf_debug(const arf_t x);
 
-void arf_print(const arf_t x);
+void arf_fprint(FILE * file, const arf_t x);
 
-void arf_printd(const arf_t y, long d);
+void arf_fprintd(FILE * file, const arf_t y, slong d);
 
-void arf_randtest(arf_t x, flint_rand_t state, long bits, long mag_bits);
+ARF_INLINE void
+arf_print(const arf_t x)
+{
+    arf_fprint(stdout, x);
+}
 
-void arf_randtest_not_zero(arf_t x, flint_rand_t state, long bits, long mag_bits);
+ARF_INLINE void
+arf_printd(const arf_t y, slong d)
+{
+    arf_fprintd(stdout, y, d);
+}
 
-void arf_randtest_special(arf_t x, flint_rand_t state, long bits, long mag_bits);
+void arf_randtest(arf_t x, flint_rand_t state, slong bits, slong mag_bits);
+
+void arf_randtest_not_zero(arf_t x, flint_rand_t state, slong bits, slong mag_bits);
+
+void arf_randtest_special(arf_t x, flint_rand_t state, slong bits, slong mag_bits);
 
 #define MUL_MPFR_MIN_LIMBS 25
 #define MUL_MPFR_MAX_LIMBS 10000
@@ -870,12 +830,12 @@ void arf_randtest_special(arf_t x, flint_rand_t state, long bits, long mag_bits)
         } \
         else if ((_xn) == 2) \
         { \
-            mp_limb_t __x1, __x0, __y1, __y0; \
-            __x0 = (_x)[0]; \
-            __x1 = (_x)[1]; \
-            __y0 = (_y)[0]; \
-            __y1 = (_y)[1]; \
-            nn_mul_2x2((_z)[3], (_z)[2], (_z)[1], (_z)[0], __x1, __x0, __y1, __y0); \
+            mp_limb_t __arb_x1, __arb_x0, __arb_y1, __arb_y0; \
+            __arb_x0 = (_x)[0]; \
+            __arb_x1 = (_x)[1]; \
+            __arb_y0 = (_y)[0]; \
+            __arb_y1 = (_y)[1]; \
+            nn_mul_2x2((_z)[3], (_z)[2], (_z)[1], (_z)[0], __arb_x1, __arb_x0, __arb_y1, __arb_y0); \
         } \
         else if ((_x) == (_y)) \
         { \
@@ -905,9 +865,9 @@ void arf_randtest_special(arf_t x, flint_rand_t state, long bits, long mag_bits)
 #define ARF_MUL_TLS_ALLOC 1000
 
 extern TLS_PREFIX mp_ptr __arf_mul_tmp;
-extern TLS_PREFIX long __arf_mul_alloc;
+extern TLS_PREFIX slong __arf_mul_alloc;
 
-extern void _arf_mul_tmp_cleanup(void);
+ARB_DLL extern void _arf_mul_tmp_cleanup(void);
 
 #define ARF_MUL_TMP_DECL \
     mp_limb_t tmp_stack[ARF_MUL_STACK_ALLOC]; \
@@ -941,11 +901,11 @@ extern void _arf_mul_tmp_cleanup(void);
 
 void arf_mul_special(arf_t z, const arf_t x, const arf_t y);
 
-int arf_mul_via_mpfr(arf_t z, const arf_t x, const arf_t y, long prec, arf_rnd_t rnd);
+int arf_mul_via_mpfr(arf_t z, const arf_t x, const arf_t y, slong prec, arf_rnd_t rnd);
 
-int arf_mul_rnd_any(arf_ptr z, arf_srcptr x, arf_srcptr y, long prec, arf_rnd_t rnd);
+int arf_mul_rnd_any(arf_ptr z, arf_srcptr x, arf_srcptr y, slong prec, arf_rnd_t rnd);
 
-int arf_mul_rnd_down(arf_ptr z, arf_srcptr x, arf_srcptr y, long prec);
+int arf_mul_rnd_down(arf_ptr z, arf_srcptr x, arf_srcptr y, slong prec);
 
 #define arf_mul(z, x, y, prec, rnd)              \
     ((rnd == FMPR_RND_DOWN)                      \
@@ -953,7 +913,7 @@ int arf_mul_rnd_down(arf_ptr z, arf_srcptr x, arf_srcptr y, long prec);
         : arf_mul_rnd_any(z, x, y, prec, rnd))
 
 ARF_INLINE int
-arf_neg_mul(arf_t z, const arf_t x, const arf_t y, long prec, arf_rnd_t rnd)
+arf_neg_mul(arf_t z, const arf_t x, const arf_t y, slong prec, arf_rnd_t rnd)
 {
     if (arf_is_special(y))
     {
@@ -971,7 +931,7 @@ arf_neg_mul(arf_t z, const arf_t x, const arf_t y, long prec, arf_rnd_t rnd)
 }
 
 ARF_INLINE int
-arf_mul_ui(arf_ptr z, arf_srcptr x, ulong y, long prec, arf_rnd_t rnd)
+arf_mul_ui(arf_ptr z, arf_srcptr x, ulong y, slong prec, arf_rnd_t rnd)
 {
     arf_t t;
     arf_init_set_ui(t, y); /* no need to free */
@@ -979,17 +939,17 @@ arf_mul_ui(arf_ptr z, arf_srcptr x, ulong y, long prec, arf_rnd_t rnd)
 }
 
 ARF_INLINE int
-arf_mul_si(arf_ptr z, arf_srcptr x, long y, long prec, arf_rnd_t rnd)
+arf_mul_si(arf_ptr z, arf_srcptr x, slong y, slong prec, arf_rnd_t rnd)
 {
     arf_t t;
     arf_init_set_si(t, y); /* no need to free */
     return arf_mul(z, x, t, prec, rnd);
 }
 
-int arf_mul_mpz(arf_ptr z, arf_srcptr x, const mpz_t y, long prec, arf_rnd_t rnd);
+int arf_mul_mpz(arf_ptr z, arf_srcptr x, const mpz_t y, slong prec, arf_rnd_t rnd);
 
 ARF_INLINE int
-arf_mul_fmpz(arf_ptr z, arf_srcptr x, const fmpz_t y, long prec, arf_rnd_t rnd)
+arf_mul_fmpz(arf_ptr z, arf_srcptr x, const fmpz_t y, slong prec, arf_rnd_t rnd)
 {
     if (!COEFF_IS_MPZ(*y))
         return arf_mul_si(z, x, *y, prec, rnd);
@@ -1001,9 +961,9 @@ arf_mul_fmpz(arf_ptr z, arf_srcptr x, const fmpz_t y, long prec, arf_rnd_t rnd)
 #define ARF_ADD_TLS_ALLOC 1000
 
 extern TLS_PREFIX mp_ptr __arf_add_tmp;
-extern TLS_PREFIX long __arf_add_alloc;
+extern TLS_PREFIX slong __arf_add_alloc;
 
-extern void _arf_add_tmp_cleanup(void);
+ARB_DLL extern void _arf_add_tmp_cleanup(void);
 
 #define ARF_ADD_TMP_DECL \
     mp_limb_t tmp_stack[ARF_ADD_STACK_ALLOC]; \
@@ -1037,24 +997,24 @@ extern void _arf_add_tmp_cleanup(void);
 
 int _arf_add_mpn(arf_t z, mp_srcptr xp, mp_size_t xn, int xsgnbit,
     const fmpz_t xexp, mp_srcptr yp, mp_size_t yn, int ysgnbit,
-    mp_bitcnt_t shift, long prec, arf_rnd_t rnd);
+    mp_bitcnt_t shift, slong prec, arf_rnd_t rnd);
 
-int arf_add(arf_ptr z, arf_srcptr x, arf_srcptr y, long prec, arf_rnd_t rnd);
-int arf_add_si(arf_ptr z, arf_srcptr x, long y, long prec, arf_rnd_t rnd);
-int arf_add_ui(arf_ptr z, arf_srcptr x, ulong y, long prec, arf_rnd_t rnd);
-int arf_add_fmpz(arf_ptr z, arf_srcptr x, const fmpz_t y, long prec, arf_rnd_t rnd);
+int arf_add(arf_ptr z, arf_srcptr x, arf_srcptr y, slong prec, arf_rnd_t rnd);
+int arf_add_si(arf_ptr z, arf_srcptr x, slong y, slong prec, arf_rnd_t rnd);
+int arf_add_ui(arf_ptr z, arf_srcptr x, ulong y, slong prec, arf_rnd_t rnd);
+int arf_add_fmpz(arf_ptr z, arf_srcptr x, const fmpz_t y, slong prec, arf_rnd_t rnd);
 
-int arf_add_fmpz_2exp(arf_ptr z, arf_srcptr x, const fmpz_t y, const fmpz_t exp, long prec, arf_rnd_t rnd);
+int arf_add_fmpz_2exp(arf_ptr z, arf_srcptr x, const fmpz_t y, const fmpz_t exp, slong prec, arf_rnd_t rnd);
 
-int arf_sub(arf_ptr z, arf_srcptr x, arf_srcptr y, long prec, arf_rnd_t rnd);
-int arf_sub_si(arf_ptr z, arf_srcptr x, long y, long prec, arf_rnd_t rnd);
-int arf_sub_ui(arf_ptr z, arf_srcptr x, ulong y, long prec, arf_rnd_t rnd);
-int arf_sub_fmpz(arf_ptr z, arf_srcptr x, const fmpz_t y, long prec, arf_rnd_t rnd);
+int arf_sub(arf_ptr z, arf_srcptr x, arf_srcptr y, slong prec, arf_rnd_t rnd);
+int arf_sub_si(arf_ptr z, arf_srcptr x, slong y, slong prec, arf_rnd_t rnd);
+int arf_sub_ui(arf_ptr z, arf_srcptr x, ulong y, slong prec, arf_rnd_t rnd);
+int arf_sub_fmpz(arf_ptr z, arf_srcptr x, const fmpz_t y, slong prec, arf_rnd_t rnd);
 
-int arf_addmul(arf_ptr z, arf_srcptr x, arf_srcptr y, long prec, arf_rnd_t rnd);
+int arf_addmul(arf_ptr z, arf_srcptr x, arf_srcptr y, slong prec, arf_rnd_t rnd);
 
 ARF_INLINE int
-arf_addmul_ui(arf_ptr z, arf_srcptr x, ulong y, long prec, arf_rnd_t rnd)
+arf_addmul_ui(arf_ptr z, arf_srcptr x, ulong y, slong prec, arf_rnd_t rnd)
 {
     arf_t t;
     arf_init_set_ui(t, y); /* no need to free */
@@ -1062,17 +1022,17 @@ arf_addmul_ui(arf_ptr z, arf_srcptr x, ulong y, long prec, arf_rnd_t rnd)
 }
 
 ARF_INLINE int
-arf_addmul_si(arf_ptr z, arf_srcptr x, long y, long prec, arf_rnd_t rnd)
+arf_addmul_si(arf_ptr z, arf_srcptr x, slong y, slong prec, arf_rnd_t rnd)
 {
     arf_t t;
     arf_init_set_si(t, y); /* no need to free */
     return arf_addmul(z, x, t, prec, rnd);
 }
 
-int arf_addmul_mpz(arf_ptr z, arf_srcptr x, const mpz_t y, long prec, arf_rnd_t rnd);
+int arf_addmul_mpz(arf_ptr z, arf_srcptr x, const mpz_t y, slong prec, arf_rnd_t rnd);
 
 ARF_INLINE int
-arf_addmul_fmpz(arf_ptr z, arf_srcptr x, const fmpz_t y, long prec, arf_rnd_t rnd)
+arf_addmul_fmpz(arf_ptr z, arf_srcptr x, const fmpz_t y, slong prec, arf_rnd_t rnd)
 {
     if (!COEFF_IS_MPZ(*y))
         return arf_addmul_si(z, x, *y, prec, rnd);
@@ -1080,10 +1040,10 @@ arf_addmul_fmpz(arf_ptr z, arf_srcptr x, const fmpz_t y, long prec, arf_rnd_t rn
         return arf_addmul_mpz(z, x, COEFF_TO_PTR(*y), prec, rnd);
 }
 
-int arf_submul(arf_ptr z, arf_srcptr x, arf_srcptr y, long prec, arf_rnd_t rnd);
+int arf_submul(arf_ptr z, arf_srcptr x, arf_srcptr y, slong prec, arf_rnd_t rnd);
 
 ARF_INLINE int
-arf_submul_ui(arf_ptr z, arf_srcptr x, ulong y, long prec, arf_rnd_t rnd)
+arf_submul_ui(arf_ptr z, arf_srcptr x, ulong y, slong prec, arf_rnd_t rnd)
 {
     arf_t t;
     arf_init_set_ui(t, y); /* no need to free */
@@ -1091,17 +1051,17 @@ arf_submul_ui(arf_ptr z, arf_srcptr x, ulong y, long prec, arf_rnd_t rnd)
 }
 
 ARF_INLINE int
-arf_submul_si(arf_ptr z, arf_srcptr x, long y, long prec, arf_rnd_t rnd)
+arf_submul_si(arf_ptr z, arf_srcptr x, slong y, slong prec, arf_rnd_t rnd)
 {
     arf_t t;
     arf_init_set_si(t, y); /* no need to free */
     return arf_submul(z, x, t, prec, rnd);
 }
 
-int arf_submul_mpz(arf_ptr z, arf_srcptr x, const mpz_t y, long prec, arf_rnd_t rnd);
+int arf_submul_mpz(arf_ptr z, arf_srcptr x, const mpz_t y, slong prec, arf_rnd_t rnd);
 
 ARF_INLINE int
-arf_submul_fmpz(arf_ptr z, arf_srcptr x, const fmpz_t y, long prec, arf_rnd_t rnd)
+arf_submul_fmpz(arf_ptr z, arf_srcptr x, const fmpz_t y, slong prec, arf_rnd_t rnd)
 {
     if (!COEFF_IS_MPZ(*y))
         return arf_submul_si(z, x, *y, prec, rnd);
@@ -1109,10 +1069,12 @@ arf_submul_fmpz(arf_ptr z, arf_srcptr x, const fmpz_t y, long prec, arf_rnd_t rn
         return arf_submul_mpz(z, x, COEFF_TO_PTR(*y), prec, rnd);
 }
 
-int arf_div(arf_ptr z, arf_srcptr x, arf_srcptr y, long prec, arf_rnd_t rnd);
+int arf_sosq(arf_t z, const arf_t x, const arf_t y, slong prec, arf_rnd_t rnd);
+
+int arf_div(arf_ptr z, arf_srcptr x, arf_srcptr y, slong prec, arf_rnd_t rnd);
 
 ARF_INLINE int
-arf_div_ui(arf_ptr z, arf_srcptr x, ulong y, long prec, arf_rnd_t rnd)
+arf_div_ui(arf_ptr z, arf_srcptr x, ulong y, slong prec, arf_rnd_t rnd)
 {
     arf_t t;
     arf_init_set_ui(t, y); /* no need to free */
@@ -1120,7 +1082,7 @@ arf_div_ui(arf_ptr z, arf_srcptr x, ulong y, long prec, arf_rnd_t rnd)
 }
 
 ARF_INLINE int
-arf_ui_div(arf_ptr z, ulong x, arf_srcptr y, long prec, arf_rnd_t rnd)
+arf_ui_div(arf_ptr z, ulong x, arf_srcptr y, slong prec, arf_rnd_t rnd)
 {
     arf_t t;
     arf_init_set_ui(t, x); /* no need to free */
@@ -1128,7 +1090,7 @@ arf_ui_div(arf_ptr z, ulong x, arf_srcptr y, long prec, arf_rnd_t rnd)
 }
 
 ARF_INLINE int
-arf_div_si(arf_ptr z, arf_srcptr x, long y, long prec, arf_rnd_t rnd)
+arf_div_si(arf_ptr z, arf_srcptr x, slong y, slong prec, arf_rnd_t rnd)
 {
     arf_t t;
     arf_init_set_si(t, y); /* no need to free */
@@ -1136,7 +1098,7 @@ arf_div_si(arf_ptr z, arf_srcptr x, long y, long prec, arf_rnd_t rnd)
 }
 
 ARF_INLINE int
-arf_si_div(arf_ptr z, long x, arf_srcptr y, long prec, arf_rnd_t rnd)
+arf_si_div(arf_ptr z, slong x, arf_srcptr y, slong prec, arf_rnd_t rnd)
 {
     arf_t t;
     arf_init_set_si(t, x); /* no need to free */
@@ -1144,7 +1106,7 @@ arf_si_div(arf_ptr z, long x, arf_srcptr y, long prec, arf_rnd_t rnd)
 }
 
 ARF_INLINE int
-arf_div_fmpz(arf_ptr z, arf_srcptr x, const fmpz_t y, long prec, arf_rnd_t rnd)
+arf_div_fmpz(arf_ptr z, arf_srcptr x, const fmpz_t y, slong prec, arf_rnd_t rnd)
 {
     arf_t t;
     int r;
@@ -1156,7 +1118,7 @@ arf_div_fmpz(arf_ptr z, arf_srcptr x, const fmpz_t y, long prec, arf_rnd_t rnd)
 }
 
 ARF_INLINE int
-arf_fmpz_div(arf_ptr z, const fmpz_t x, arf_srcptr y, long prec, arf_rnd_t rnd)
+arf_fmpz_div(arf_ptr z, const fmpz_t x, arf_srcptr y, slong prec, arf_rnd_t rnd)
 {
     arf_t t;
     int r;
@@ -1168,7 +1130,7 @@ arf_fmpz_div(arf_ptr z, const fmpz_t x, arf_srcptr y, long prec, arf_rnd_t rnd)
 }
 
 ARF_INLINE int
-arf_fmpz_div_fmpz(arf_ptr z, const fmpz_t x, const fmpz_t y, long prec, arf_rnd_t rnd)
+arf_fmpz_div_fmpz(arf_ptr z, const fmpz_t x, const fmpz_t y, slong prec, arf_rnd_t rnd)
 {
     arf_t t, u;
     int r;
@@ -1182,15 +1144,15 @@ arf_fmpz_div_fmpz(arf_ptr z, const fmpz_t x, const fmpz_t y, long prec, arf_rnd_
     return r;
 }
 
-int arf_sqrt(arf_ptr z, arf_srcptr x, long prec, arf_rnd_t rnd);
+int arf_sqrt(arf_ptr z, arf_srcptr x, slong prec, arf_rnd_t rnd);
 
-int arf_sqrt_ui(arf_t z, ulong x, long prec, arf_rnd_t rnd);
+int arf_sqrt_ui(arf_t z, ulong x, slong prec, arf_rnd_t rnd);
 
-int arf_sqrt_fmpz(arf_t z, const fmpz_t x, long prec, arf_rnd_t rnd);
+int arf_sqrt_fmpz(arf_t z, const fmpz_t x, slong prec, arf_rnd_t rnd);
 
-int arf_rsqrt(arf_ptr z, arf_srcptr x, long prec, arf_rnd_t rnd);
+int arf_rsqrt(arf_ptr z, arf_srcptr x, slong prec, arf_rnd_t rnd);
 
-int arf_root(arf_t z, const arf_t x, ulong k, long prec, arf_rnd_t rnd);
+int arf_root(arf_t z, const arf_t x, ulong k, slong prec, arf_rnd_t rnd);
 
 /* Magnitude bounds */
 
@@ -1247,18 +1209,18 @@ mag_fast_init_set_arf(mag_t y, const arf_t x)
 }
 
 ARF_INLINE void
-arf_mag_fast_add_ulp(mag_t z, const mag_t x, const arf_t y, long prec)
+arf_mag_fast_add_ulp(mag_t z, const mag_t x, const arf_t y, slong prec)
 {
     mag_fast_add_2exp_si(z, x, ARF_EXP(y) - prec);
 }
 
 ARF_INLINE void
-arf_mag_add_ulp(mag_t z, const mag_t x, const arf_t y, long prec)
+arf_mag_add_ulp(mag_t z, const mag_t x, const arf_t y, slong prec)
 {
     if (ARF_IS_SPECIAL(y))
     {
-        printf("error: ulp error not defined for special value!\n");
-        abort();
+        flint_printf("error: ulp error not defined for special value!\n");
+        flint_abort();
     }
     else if (MAG_IS_LAGOM(z) && MAG_IS_LAGOM(x) && ARF_IS_LAGOM(y))
     {
@@ -1275,12 +1237,12 @@ arf_mag_add_ulp(mag_t z, const mag_t x, const arf_t y, long prec)
 }
 
 ARF_INLINE void
-arf_mag_set_ulp(mag_t z, const arf_t y, long prec)
+arf_mag_set_ulp(mag_t z, const arf_t y, slong prec)
 {
     if (ARF_IS_SPECIAL(y))
     {
-        printf("error: ulp error not defined for special value!\n");
-        abort();
+        flint_printf("error: ulp error not defined for special value!\n");
+        flint_abort();
     }
     else
     {
@@ -1292,28 +1254,39 @@ arf_mag_set_ulp(mag_t z, const arf_t y, long prec)
 void arf_get_fmpq(fmpq_t y, const arf_t x);
 
 ARF_INLINE int
-arf_set_fmpq(arf_t y, const fmpq_t x, long prec, arf_rnd_t rnd)
+arf_set_fmpq(arf_t y, const fmpq_t x, slong prec, arf_rnd_t rnd)
 {
     return arf_fmpz_div_fmpz(y, fmpq_numref(x), fmpq_denref(x), prec, rnd);
 }
 
 int arf_complex_mul(arf_t e, arf_t f, const arf_t a, const arf_t b,
                                       const arf_t c, const arf_t d,
-                                      long prec, arf_rnd_t rnd);
+                                      slong prec, arf_rnd_t rnd);
 
 int arf_complex_mul_fallback(arf_t e, arf_t f,
         const arf_t a, const arf_t b,
         const arf_t c, const arf_t d,
-        long prec, arf_rnd_t rnd);
+        slong prec, arf_rnd_t rnd);
 
 int arf_complex_sqr(arf_t e, arf_t f, const arf_t a, const arf_t b,
-                                      long prec, arf_rnd_t rnd);
+                                      slong prec, arf_rnd_t rnd);
 
-int arf_sum(arf_t s, arf_srcptr terms, long len, long prec, arf_rnd_t rnd);
+int arf_sum(arf_t s, arf_srcptr terms, slong len, slong prec, arf_rnd_t rnd);
 
 double arf_get_d(const arf_t x, arf_rnd_t rnd);
 void arf_set_d(arf_t x, double v);
 
+ARF_INLINE slong
+arf_allocated_bytes(const arf_t x)
+{
+    slong size = fmpz_allocated_bytes(ARF_EXPREF(x));
+
+    if (ARF_HAS_PTR(x))
+        size += ARF_PTR_ALLOC(x) * sizeof(mp_limb_t);
+
+    return size;
+}
+
 #ifdef __cplusplus
 }
 #endif
